/***************************************************************************
                          mmx_idct.c  -  description
                             -------------------
   MMX version of the IDCT function. Suitable for all processors (PMMX, P2,
   and later) that support the MMX instruction set.
     Operations per row  : 16 PMADDWDs, 18 PADDD/PORs, 8 PSxxxs
     Operations per 4cols: 22 PMULHWs,  33 PADDSWs,    8 PSRAWs
     Operations per block:172,         210,           80
                             -------------------
    begin                : Tue Nov 16 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define shift_value  15
#define row_shift    (shift_value-4)
#define row_error    ((1 << (row_shift-1)) + (1 << (row_shift-3)))
#define col_shift    (shift_value+6-16)
#define col_error    ((1 << (col_shift-1+16)) / cosp4_16)

/* cospk_16r = (s32)( ( cos(PI*k / 16) * (1 << shift_value)) + 0.5 ) */
#define cosp1_16 ( 32138)
#define cosp2_16 ( 30274)
#define cosp3_16 ( 27246)
#define cosp4_16 ( 23170)
#define cosp5_16 ( 18205)
#define cosp6_16 ( 12540)
#define cosp7_16 (  6393)

/* cosmk_16 = (s32)( (-cos(PI*k / 16) * (1 << shift_value)) + 0.5 ) */
#define cosm1_16 (-32137)
#define cosm2_16 (-30273)
#define cosm3_16 (-27245)
#define cosm4_16 (-23169)
#define cosm5_16 (-18204)
#define cosm6_16 (-12539)
#define cosm7_16 ( -6392)

#define r_const(w3, w2, w1, w0)  w0, w1, w2, w3
#define c_const(w)               w, w, w, w
static const s16 row_const[68] = {
   /*      t1+t5                 t0+t4 */
   r_const( cosm7_16,  cosp6_16,  cosp1_16,  cosp4_16),
   r_const( cosm5_16, -cosp2_16,  cosp5_16,  cosp4_16),
   r_const( cosp3_16,  cosp4_16,  cosp3_16,  cosp2_16),
   r_const( cosm1_16, -cosp4_16,  cosp7_16,  cosp6_16),

   /*      t3+t7                 t2+t6 */
   r_const( cosm5_16, -cosp2_16,  cosp5_16,  cosp4_16),
   r_const( cosm1_16, -cosp6_16,  cosp7_16, -cosp4_16),
   r_const( cosp7_16,  cosp4_16,  cosm1_16, -cosp6_16),
   r_const( cosp3_16,  cosp4_16,  cosp3_16,  cosp2_16),

   /*      t2-t6                 t3-t7 */
   r_const(-cosm1_16, -cosp6_16, -cosp7_16,  cosp4_16),
   r_const(-cosp3_16,  cosp2_16, -cosp3_16,  cosp4_16),
   r_const(-cosp5_16,  cosp4_16, -cosm5_16, -cosp2_16),
   r_const(-cosp7_16, -cosp4_16, -cosm1_16, -cosp6_16),

   /*      t0-t4                 t1-t5 */
   r_const(-cosp3_16,  cosp2_16, -cosp3_16,  cosp4_16),
   r_const(-cosp7_16,  cosp6_16, -cosm1_16, -cosp4_16),
   r_const(-cosp1_16,  cosp4_16, -cosm7_16,  cosp6_16),
   r_const(-cosp5_16,  cosp4_16, -cosm5_16, -cosp2_16),

   /* rounding values */
   r_const(0,         row_error,         0, row_error)
},
                 col_const[40] = {
   c_const( cosp4_16),
   c_const( cosp2_16),
   c_const( cosp6_16),
   c_const(-cosp6_16),
   c_const( cosp1_16),
   c_const( cosp3_16),
   c_const(-cosp5_16),
   c_const( cosp7_16),

   /* rounding values */
   c_const(col_error),
   c_const(1)
};

/* Does an 8x8 2D iDCT operation.
      INPUT: data - An array holding the 64 dct coefficients.
      OUTPUT: data - The 64 dct coefficients are replaced by the 64
                       recovered samples. */
#ifdef _IEEE_TEST_
void j_rev_dct(DCTBLOCK data)
#else
void idct(s16 *data)
#endif
{
#if defined(__GCC__)

/**** process rows ****/
   __asm__ __volatile__(
    /* get input from memory, and put it in order
         registers used: %0 = data
                         %%mm0, %%mm1, %%mm2, %%mm3 */
    /* op_code  src      dest */
      "IDCT_ROW_LOOP:\n\t"
      "movq "  "(%0), " "%%mm0\n\t" /*P2  mm0 = data[3] data[2] data[1] data[0] */
      "movq "  "%%mm0, ""%%mm2\n\t" /*P01 */
      "movq "  "8(%0), ""%%mm1\n\t" /*P2  mm1 = data[7] data[6] data[5] data[4] */
      "movq "  "%%mm0, ""%%mm4\n\t" /*P01 */
      "psllq " "$32, "  "%%mm2\n\t" /*P1 */
      "movq "  "%%mm1, ""%%mm3\n\t" /*P01 */
      "psrlq " "$32, "  "%%mm4\n\t" /*P1 */
      "por "   "%%mm4, ""%%mm2\n\t" /*P01 mm2 = data[1] data[0] data[3] data[2] */
      "psllq " "$32, "  "%%mm3\n\t" /*P1 */
      "movq "  "%%mm1, ""%%mm5\n\t" /*P01 */
      "psrlq " "$32, "  "%%mm5\n\t" /*P1 */
      "por "   "%%mm5, ""%%mm3\n\t" /*P01 mm3 = data[5] data[4] data[7] data[6] */

    /* calculate t0+t4, and t1+t5
         registers used: %1 = row_const
                         %2 = row_shift
                         %%mm4 */
    /* op_code     src         dest */
      "movq "     "(%1), "    "%%mm4\n\t" /*P2 */
      "pmaddwd "  "%%mm0, "   "%%mm4\n\t" /*P0 */
      "movq "     "8(%1), "   "%%mm5\n\t" /*P2 */
      "pmaddwd "  "%%mm1, "   "%%mm5\n\t" /*P0 */
      "paddd "    "%%mm5, "   "%%mm4\n\t" /*P01 */
      "movq "     "16(%1), "  "%%mm5\n\t" /*P2 */
      "pmaddwd "  "%%mm2, "   "%%mm5\n\t" /*P0 */
      "movq "     "24(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm3, "   "%%mm6\n\t" /*P0 */
      "paddd "    "128(%1), " "%%mm4\n\t" /*P2 P01 add row_error */
      "paddd "    "%%mm6, "   "%%mm5\n\t" /*P01 */
      "paddd "    "%%mm5, "   "%%mm4\n\t" /*P01 */
      "psrad "    "%2, "      "%%mm4\n\t" /*P1  mm4 = (t1+t5) (t0+t4) */

    /* calculate t2+t6, and t3+t7
       pack t0+t4, t1+t5, t2+t6, and t3+t7
         registers used: %1 = row_const
                         %%mm4 */
    /* op_code     src         dest */
      "movq "     "32(%1), "  "%%mm5\n\t" /*P2 */
      "pmaddwd "  "%%mm0, "   "%%mm5\n\t" /*P0 */
      "movq "     "40(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm1, "   "%%mm6\n\t" /*P0 */
      "paddd "    "%%mm6, "   "%%mm5\n\t" /*P01 */
      "movq "     "48(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm2, "   "%%mm6\n\t" /*P0 */
      "movq "     "56(%1), "  "%%mm7\n\t" /*P2 */
      "pmaddwd "  "%%mm3, "   "%%mm7\n\t" /*P0 */
      "paddd "    "128(%1), " "%%mm5\n\t" /*P2 P01 add row_error */
      "paddd "    "%%mm7, "   "%%mm6\n\t" /*P01 */
      "paddd "    "%%mm6, "   "%%mm5\n\t" /*P01 */
      "psrad "    "%2, "      "%%mm5\n\t" /*P1  mm5 = (t3+t7) (t2+t6) */
      "packssdw " "%%mm5, "   "%%mm4\n\t" /*P1  mm4 = (t3+t7) (t2+t6) (t1+t5) (t0+t4) */

    /* calculate t3-t7, and t2-t6
         registers used: %1 = row_const
                         %%mm5 */
    /* op_code     src         dest */
      "movq "     "64(%1), "  "%%mm5\n\t" /*P2 */
      "pmaddwd "  "%%mm0, "   "%%mm5\n\t" /*P0 */
      "movq "     "72(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm1, "   "%%mm6\n\t" /*P0 */
      "paddd "    "%%mm6, "   "%%mm5\n\t" /*P01 */
      "movq "     "80(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm2, "   "%%mm6\n\t" /*P0 */
      "movq "     "88(%1), "  "%%mm7\n\t" /*P2 */
      "pmaddwd "  "%%mm3, "   "%%mm7\n\t" /*P0 */
      "paddd "    "128(%1), " "%%mm5\n\t" /*P2 P01 add row_error */
      "paddd "    "%%mm7, "   "%%mm6\n\t" /*P01 */
      "paddd "    "%%mm6, "   "%%mm5\n\t" /*P01 */
      "psrad "    "%2,  "     "%%mm5\n\t" /*P1  mm5 = (t2-t6) (t3-t7) */

    /* calculate t1-t5, and t0-t4
       pack t3-t7, t2-t6, t1-t5, and t0-t4
         registers used: %%mm5
         registers freed: %%mm0, %%mm1, %%mm2, %%mm3, %%mm4, %%mm5 */
    /* op_code     src         dest */
      "movq "     "96(%1), "  "%%mm6\n\t" /*P2 */
      "pmaddwd "  "%%mm0, "   "%%mm6\n\t" /*P0    free mm0 */
      "movq "     "104(%1), " "%%mm7\n\t" /*P2 */
      "pmaddwd "  "%%mm1, "   "%%mm7\n\t" /*P0    free mm1 */
      "paddd "    "%%mm7, "   "%%mm6\n\t" /*P01 */
      "movq "     "112(%1), " "%%mm7\n\t" /*P2 */
      "pmaddwd "  "%%mm2, "   "%%mm7\n\t" /*P0    free mm2 */
      "movq "     "120(%1), " "%%mm0\n\t" /*P2 */
      "pmaddwd "  "%%mm3, "   "%%mm0\n\t" /*P0    free mm3 */
      "movq "     "%%mm4, "   "(%0)\n\t"  /*P3 P4 free mm4 */
      "paddd "    "128(%1), " "%%mm6\n\t" /*P2 P01 add row_error */
      "paddd "    "%%mm0, "   "%%mm7\n\t" /*P01 */
      "paddd "    "%%mm7, "   "%%mm6\n\t" /*P01 */
      "psrad "    "%2, "      "%%mm6\n\t" /*P1    mm6 = (t0-t4) (t1-t5) */
      "packssdw " "%%mm6, "   "%%mm5\n\t" /*P1    mm5 = (t0-t4) (t1-t5) (t2-t6) (t3-t7) */
      "movq "     "%%mm5, "   "8(%0)\n\t" /*P3 P4 free mm5 */

      "addl "     "$16, "     "%0\n\t"    /*P01 data += 8; */
      "dec "                  "%3\n\t"    /*P01 loop counter */
      "jnz "      "IDCT_ROW_LOOP\n\t"
      "subl "     "$128, "    "%0"        /* avoid optimization problems */

   /* output list */
   :

   /* input list */
   : "q" (data),
     "q" (row_const),
     "I" (row_shift),
     "r" ((int)8)
   );

/**** process cols ****/
   __asm__ __volatile__(
    /* calculate t0, t1, t2, t3
         registers used: %0 = data
                         %1 = col_const
                         %%mm0, %%mm1, %%mm2, %%mm3 */
    /* op_code  src         dest */
      "IDCT_COL_LOOP:\n\t"
      "movq "  "(%0),   "  "%%mm7\n\t" /*P2  mm7 = x0 */
      "movq "  "32(%0), "  "%%mm5\n\t" /*P2  mm5 = x2 */
      "movq "  "64(%0), "  "%%mm0\n\t" /*P2  mm0 = x4 */
      "movq "  "96(%0), "  "%%mm2\n\t" /*P2  mm2 = x6 */

      "paddsw ""64(%1), "  "%%mm7\n\t" /*P2 P01 add col_error */
      
      "movq "  "(%1), "    "%%mm1\n\t" /*P2  mm1 = cosp4_16 */
      "movq "  "%%mm7, "   "%%mm6\n\t" /*P01 mm6 = x0 */
      "paddsw ""%%mm0, "   "%%mm7\n\t" /*P01 mm7 = x0+x4 */
      "psubsw ""%%mm0, "   "%%mm6\n\t" /*P01 mm6 = x0-x4 */
      "pmulhw ""%%mm1, "   "%%mm7\n\t" /*P0  mm7 = cosp4_16 * (x0+x4) */
      "pmulhw ""%%mm1, "   "%%mm6\n\t" /*P0  mm6 = cosp4_16 * (x0-x4) */

      "movq "  "8(%1), "   "%%mm0\n\t" /*P2  mm0 = cosp2_16 */
      "movq "  "%%mm5, "   "%%mm4\n\t" /*P01 mm4 = x2 */
      "movq "  "%%mm2, "   "%%mm3\n\t" /*P01 mm3 = x6 */
      "pmulhw ""%%mm0, "   "%%mm5\n\t" /*P0  mm5 = cosp2_16 * x2 */
      "pmulhw ""%%mm0, "   "%%mm2\n\t" /*P0  mm2 = cosp2_16 * x6 */

      "movq "  "16(%1), "  "%%mm1\n\t" /*P2  mm1 =  cosp6_16 */
      "movq "  "24(%1), "  "%%mm0\n\t" /*P2  mm0 = -cosp6_16 */
      "pmulhw ""%%mm1, "   "%%mm4\n\t" /*P0  mm4 =  cosp6_16 * x2 */
      "pmulhw ""%%mm0, "   "%%mm3\n\t" /*P0  mm3 = -cosp6_16 * x6 */

      "movq "  "%%mm7, "   "%%mm0\n\t" /*P01 */
      "psubsw ""%%mm3, "   "%%mm0\n\t" /*P01 mm0 = t7 - t3 */
      "paddsw ""%%mm7, "   "%%mm3\n\t" /*P01 mm3 = t7 + t3 */
      "movq "  "%%mm6, "   "%%mm1\n\t" /*P01 */
      "psubsw ""%%mm2, "   "%%mm1\n\t" /*P01 mm1 = t6 - t2 */
      "paddsw ""%%mm6, "   "%%mm2\n\t" /*P01 mm2 = t6 + t2 */
      "paddsw ""%%mm5, "   "%%mm0\n\t" /*P01 mm0 = t0 = t7 + t5 - t3 */
      "psubsw ""%%mm5, "   "%%mm3\n\t" /*P01 mm3 = t3 = t7 - t5 + t3 */
      "paddsw ""%%mm4, "   "%%mm1\n\t" /*P01 mm1 = t1 = t6 + t4 - t2 */
      "psubsw ""%%mm4, "   "%%mm2\n\t" /*P01 mm2 = t2 = t6 - t4 + t2 */
      
    /* calculate and store t0+t4, and t0-t4
         registers used: %0 = data
                         %1 = col_const
                         %2 = col_shift
                         %%mm6
         registers freed: %%mm0 */
    /* op_code  src         dest */
      "movq "  "16(%0), "  "%%mm4\n\t" /*P2    mm4 = x1 */
      "movq "  "48(%0), "  "%%mm5\n\t" /*P2    mm5 = x3 */
      "movq "  "80(%0), "  "%%mm6\n\t" /*P2    mm6 = x5 */
      "movq "  "112(%0), " "%%mm7\n\t" /*P2    mm7 = x7 */
      "pmulhw ""32(%1), "  "%%mm4\n\t" /*P2 P0 mm4 =  cosp1_16 * x1 */
      "pmulhw ""40(%1), "  "%%mm5\n\t" /*P2 P0 mm5 =  cosp3_16 * x3 */
      "paddsw ""%%mm5, "   "%%mm4\n\t" /*P01 */
      "pmulhw ""48(%1), "  "%%mm6\n\t" /*P2 P0 mm6 = -cosp5_16 * x5 */
      "psubsw ""%%mm6, "   "%%mm4\n\t" /*P01 */
      "movq "  "%%mm7, "   "%%mm6\n\t" /*P01   mm6 = x7 */
      "pmulhw ""56(%1), "  "%%mm7\n\t" /*P2 P0 mm7 =  cosp7_16 * x7 */
      "paddsw ""%%mm7, "   "%%mm4\n\t" /*P01   mm4 = t4 */
      "paddsw ""72(%1), "  "%%mm4\n\t" /*P2 P01 reduce rounding error */
      "movq "  "%%mm4, "   "%%mm5\n\t" /*P01 */
      "paddsw ""%%mm0, "   "%%mm4\n\t" /*P01   mm4 = t0+t4 */
      "psubsw ""%%mm5, "   "%%mm0\n\t" /*P01   mm0 = t0-t4 */

      "psraw " "%2, "      "%%mm4\n\t"    /*P1 */
      "movq "  "%%mm4, "   "(%0)\n\t"     /*P3 P4 */
      "psraw " "%2, "      "%%mm0\n\t"    /*P1 */
      "movq "  "%%mm0, "   "112(%0)\n\t"  /*P3 P4 free mm0 */

    /* calculate and store t1+t5, and t1-t5
         registers used: %0 = data
                         %1 = col_const
                         %2 = col_shift
                         %%mm5
         registers freed: %%mm1 */
    /* op_code  src         dest */
      "movq "  "%%mm6, "   "%%mm7\n\t" /*P01   mm7 = x7 */
      "movq "  "80(%0), "  "%%mm0\n\t" /*P2    mm0 = x5 */
      "movq "  "48(%0), "  "%%mm5\n\t" /*P2    mm5 = x3 */
      "movq "  "16(%0), "  "%%mm4\n\t" /*P2    mm4 = x1 */
      "pmulhw ""48(%1), "  "%%mm7\n\t" /*P2 P0 mm7 = -cosp5_16 * x7 */
      "pmulhw ""32(%1), "  "%%mm0\n\t" /*P2 P0 mm0 =  cosp1_16 * x5 */
      "psubsw ""%%mm0, "   "%%mm7\n\t" /*P01 */
      "pmulhw ""56(%1), "  "%%mm5\n\t" /*P2 P0 mm5 =  cosp7_16 * x3 */
      "psubsw ""%%mm5, "   "%%mm7\n\t" /*P01 */
      "movq "  "%%mm4, "   "%%mm5\n\t" /*P01   mm5 = x1 */
      "pmulhw ""40(%1), "  "%%mm4\n\t" /*P2 P0 mm4 =  cosp3_16 * x1 */
      "paddsw ""%%mm4, "   "%%mm7\n\t" /*P01   mm7 = t5 */
      "movq "  "%%mm7, "   "%%mm0\n\t"
      "paddsw ""%%mm1, "   "%%mm7\n\t" /*P01   mm7 = t1+t5 */
      "psubsw ""%%mm0, "   "%%mm1\n\t" /*P01   mm1 = t1-t5 */

      "psraw " "%2, "      "%%mm7\n\t" /*P1 */
      "movq "  "%%mm7, "   "16(%0)\n\t"/*P3 P4 */
      "psraw " "%2, "      "%%mm1\n\t" /*P1 */
      "movq "  "%%mm1, "   "96(%0)\n\t"/*P3 P4 free mm1 */

    /* calculate and store t2+t6, and t2-t6
         registers used: %0 = data
                         %1 = col_const
                         %2 = col_shift
                         %%mm1
         registers freed: %%mm2 */
    /* op_code  src         dest */
      "movq "  "%%mm6, "   "%%mm7\n\t" /*P01   mm7 = x7 */
      "movq "  "%%mm5, "   "%%mm4\n\t" /*P01   mm4 = x1 */
      "movq "  "48(%0), "  "%%mm1\n\t" /*P2    mm1 = x3 */
      "movq "  "80(%0), "  "%%mm0\n\t" /*P2    mm0 = x5 */
      "pmulhw ""40(%1), "  "%%mm7\n\t" /*P2 P0 mm7 =  cosp3_16 * x7 */
      "pmulhw ""48(%1), "  "%%mm4\n\t" /*P2 P0 mm4 = -cosp5_16 * x1 */
      "psubsw ""%%mm4, "   "%%mm7\n\t" /*P01 */
      "pmulhw ""32(%1), "  "%%mm1\n\t" /*P2 P0 mm1 =  cosp1_16 * x3 */
      "psubsw ""%%mm1, "   "%%mm7\n\t" /*P01 */
      "movq "  "%%mm0, "   "%%mm1\n\t" /*P01   mm1 = x5 */
      "pmulhw ""56(%1), "  "%%mm0\n\t" /*P2 P0 mm0 =  cosp7_16 * x5 */
      "paddsw ""%%mm0, "   "%%mm7\n\t" /*P01   mm7 = t6 */
      "movq "  "%%mm7, "   "%%mm4\n\t" /*P01 */
      "paddsw ""%%mm2, "   "%%mm7\n\t" /*P01   mm7 = t2+t6 */
      "psubsw ""%%mm4, "   "%%mm2\n\t" /*P01   mm2 = t2-t6 */

      "psraw " "%2, "      "%%mm7\n\t" /*P1 */
      "movq "  "%%mm7, "   "32(%0)\n\t"/*P3 P4 */
      "psraw " "%2, "      "%%mm2\n\t" /*P1 */
      "movq "  "%%mm2, "   "80(%0)\n\t"/*P3 P4 free mm2 */

    /* calculate and store t3+t7, and t3-t7
         registers used: %0 = data
                         %1 = col_const
                         %2 = col_shift
         registers freed: %%mm1, %%mm3, %%mm5, %%mm6 */
    /* op_code  src         dest */
      "movq "  "48(%0), "  "%%mm4\n\t" /*P2    mm4 = x3 */
      "pmulhw ""56(%1), "  "%%mm5\n\t" /*P2 P0 mm5 =  cosp7_16 * x1 */
      "pmulhw ""48(%1), "  "%%mm4\n\t" /*P2 P0 mm4 = -cosp5_16 * x3 */
      "paddsw ""%%mm4, "   "%%mm5\n\t" /*P01 */
      "pmulhw ""40(%1), "  "%%mm1\n\t" /*P2 P0 mm1 =  cosp3_16 * x5 */
      "paddsw ""%%mm1, "   "%%mm5\n\t" /*P01 */
      "pmulhw ""32(%1), "  "%%mm6\n\t" /*P2 P0 mm6 =  cosp1_16 * x7 */
      "psubsw ""%%mm6, "   "%%mm5\n\t" /*P01   mm5 = t7 */
      "paddsw ""72(%1), "  "%%mm5\n\t" /*P2 P01 reduce rounding error */
      "movq "  "%%mm5, "   "%%mm4\n\t" /*P01 */
      "paddsw ""%%mm3, "   "%%mm5\n\t" /*P01   mm5 = t3+t7 */
      "psubsw ""%%mm4, "   "%%mm3\n\t" /*P01   mm3 = t3-t7 */

      "psraw " "%2, "      "%%mm5\n\t" /*P1 */
      "movq "  "%%mm5, "   "48(%0)\n\t"/*P3 P4 */
      "psraw " "%2, "      "%%mm3\n\t" /*P1 */
      "movq "  "%%mm3, "   "64(%0)\n\t"/*P3 P4 free mm3 */

      "addl "  "$8, "      "%0\n\t"    /*P01   data += 4; */
      "dec "               "%3\n\t"    /*P01   loop counter */
      "jnz "   "IDCT_COL_LOOP\n\t"
      "emms"

   /* output list */
   :

   /* input list */
   : "q" (data),
     "q" (col_const),
     "I" (col_shift),
     "r" (2)
   );

#elif defined(__MSVC__)

   __asm {
/**** process rows ****/
   /* op_code  dest  src */
      mov      eax,  dword ptr[data]
      mov      ebx,  offset row_const
      mov      edx,  offset col_const
      mov      ecx,  8

   /* get input from memory, and put it in order
         registers used: eax = data
                         mm0, mm1, mm2, mm3 */
   /* op_code  dest  src */
   IDCT_ROW_LOOP:
      movq     mm0,  [eax]    /*P2  mm0 = data[3] data[2] data[1] data[0] */
      movq     mm2,  mm0      /*P01 */
      movq     mm1,  [eax+8]  /*P2  mm1 = data[7] data[6] data[5] data[4] */
      movq     mm4,  mm0      /*P01 */
      psllq    mm2,  32       /*P1 */
      movq     mm3,  mm1      /*P01 */
      psrlq    mm4,  32       /*P1 */
      por      mm2,  mm4      /*P01 mm2 = data[1] data[0] data[3] data[2] */
      psllq    mm3,  32       /*P1 */
      movq     mm5,  mm1      /*P01 */
      psrlq    mm5,  32       /*P1 */
      por      mm3,  mm5      /*P01 mm3 = data[5] data[4] data[7] data[6] */

   /* calculate t0+t4, and t1+t5
         registers used: ebx = row_const
                         mm4 */
   /* op_code  dest  src */
      movq     mm4,  [ebx]       /*P2 */
      pmaddwd  mm4,  mm0         /*P0 */
      movq     mm5,  [ebx+8]     /*P2 */
      pmaddwd  mm5,  mm1         /*P0 */
      paddd    mm4,  mm5         /*P01 */
      movq     mm5,  [ebx+16]    /*P2 */
      pmaddwd  mm5,  mm2         /*P0 */
      movq     mm6,  [ebx+24]    /*P2 */
      pmaddwd  mm6,  mm3         /*P0 */
      paddd    mm4,  [ebx+128]   /*P2 P01 add row_error */
      paddd    mm5,  mm6         /*P01 */
      paddd    mm4,  mm5         /*P01 */
      psrad    mm4,  row_shift   /*P1  mm4 = (t1+t5) (t0+t4) */

   /* calculate t2+t6, and t3+t7
      pack t0+t4, t1+t5, t2+t6, and t3+t7
         registers used: ebx = row_const
                         mm4 */
   /* op_code  dest  src */
      movq     mm5,  [ebx+32]    /*P2 */
      pmaddwd  mm5,  mm0         /*P0 */
      movq     mm6,  [ebx+40]    /*P2 */
      pmaddwd  mm6,  mm1         /*P0 */
      paddd    mm5,  mm6         /*P01 */
      movq     mm6,  [ebx+48]    /*P2 */
      pmaddwd  mm6,  mm2         /*P0 */
      movq     mm7,  [ebx+56]    /*P2 */
      pmaddwd  mm7,  mm3         /*P0 */
      paddd    mm5,  [ebx+128]   /*P2 P01 add row_error */
      paddd    mm6,  mm7         /*P01 */
      paddd    mm5,  mm6         /*P01 */
      psrad    mm5,  row_shift   /*P1  mm5 = (t3+t7) (t2+t6) */
      packssdw mm4,  mm5         /*P1  mm4 = (t3+t7) (t2+t6) (t1+t5) (t0+t4) */

   /* calculate t3-t7, and t2-t6
         registers used: ebx = row_const
                         mm5 */
   /* op_code  dest  src */
      movq     mm5,  [ebx+64]    /*P2 */
      pmaddwd  mm5,  mm0         /*P0 */
      movq     mm6,  [ebx+72]    /*P2 */
      pmaddwd  mm6,  mm1         /*P0 */
      paddd    mm5,  mm6         /*P01 */
      movq     mm6,  [ebx+80]    /*P2 */
      pmaddwd  mm6,  mm2         /*P0 */
      movq     mm7,  [ebx+88]    /*P2 */
      pmaddwd  mm7,  mm3         /*P0 */
      paddd    mm5,  [ebx+128]   /*P2 P01 add row_error */
      paddd    mm6,  mm7         /*P01 */
      paddd    mm5,  mm6         /*P01 */
      psrad    mm5,  row_shift   /*P1  mm5 = (t2-t6) (t3-t7) */

   /* calculate t1-t5, and t0-t4
      pack t3-t7, t2-t6, t1-t5, and t0-t4
         registers used: eax = data
                         ebx = row_const
                         mm5
         registers freed: mm0, mm1, mm2, mm3, mm4, mm5 */
   /* op_code  dest     src */
      movq     mm6,     [ebx+96]    /*P2 */
      pmaddwd  mm6,     mm0         /*P0    free mm0 */
      movq     mm7,     [ebx+104]   /*P2 */
      pmaddwd  mm7,     mm1         /*P0    free mm1 */
      paddd    mm6,     mm7         /*P01 */
      movq     mm7,     [ebx+112]   /*P2 */
      pmaddwd  mm7,     mm2         /*P0    free mm2 */
      movq     mm0,     [ebx+120]   /*P2 */
      pmaddwd  mm0,     mm3         /*P0    free mm3 */
      movq     [eax],   mm4         /*P3 P4 free mm4 */
      paddd    mm6,     [ebx+128]   /*P2 P01 add row_error */
      paddd    mm7,     mm0         /*P01 */
      paddd    mm6,     mm7         /*P01 */
      psrad    mm6,     row_shift   /*P1    mm6 = (t0-t4) (t1-t5) */
      packssdw mm5,     mm6         /*P1    mm5 = (t0-t4) (t1-t5) (t2-t6) (t3-t7) */
      movq     [eax+8], mm5         /*P3 P4 free mm5 */

      add      eax,     16          /*P01 data += 8; */
      dec      ecx                  /*P01 loop counter */
      jnz      IDCT_ROW_LOOP
      sub      eax,     128         /* reset pointer */
      mov      ecx,     2

/**** process cols ****/
   /* calculate t0, t1, t2, t3
         registers used: eax = data
                         edx = col_const
                         mm0, mm1, mm2, mm3 */
   /* op_code  dest  src */
   IDCT_COL_LOOP:
      movq     mm7,  [eax]    /*P2  mm7 = x0 */
      movq     mm5,  [eax+32] /*P2  mm5 = x2 */
      movq     mm0,  [eax+64] /*P2  mm0 = x4 */
      movq     mm2,  [eax+96] /*P2  mm2 = x6 */

      paddsw   mm7,  [edx+64] /*P2 P01 add col_error */
      
      movq     mm1,  [edx]    /*P2  mm1 = cosp4_16 */
      movq     mm6,  mm7      /*P01 mm6 = x0 */
      paddsw   mm7,  mm0      /*P01 mm7 = x0+x4 */
      psubsw   mm6,  mm0      /*P01 mm6 = x0-x4 */
      pmulhw   mm7,  mm1      /*P0  mm7 = cosp4_16 * (x0+x4) */
      pmulhw   mm6,  mm1      /*P0  mm6 = cosp4_16 * (x0-x4) */

      movq     mm0,  [edx+8]  /*P2  mm0 = cosp2_16 */
      movq     mm4,  mm5      /*P01 mm4 = x2 */
      movq     mm3,  mm2      /*P01 mm3 = x6 */
      pmulhw   mm5,  mm0      /*P0  mm5 = cosp2_16 * x2 */
      pmulhw   mm2,  mm0      /*P0  mm2 = cosp2_16 * x6 */

      movq     mm1,  [edx+16] /*P2  mm1 =  cosp6_16 */
      movq     mm0,  [edx+24] /*P2  mm0 = -cosp6_16 */
      pmulhw   mm4,  mm1      /*P0  mm4 =  cosp6_16 * x2 */
      pmulhw   mm3,  mm0      /*P0  mm3 = -cosp6_16 * x6 */

      movq     mm0,  mm7      /*P01 */
      psubsw   mm0,  mm3      /*P01 mm0 = t7 - t3 */
      paddsw   mm3,  mm7      /*P01 mm3 = t7 + t3 */
      movq     mm1,  mm6      /*P01 */
      psubsw   mm1,  mm2      /*P01 mm1 = t6 - t2 */
      paddsw   mm2,  mm6      /*P01 mm2 = t6 + t2 */
      paddsw   mm0,  mm5      /*P01 mm0 = t0 = t7 + t5 - t3 */
      psubsw   mm3,  mm5      /*P01 mm3 = t3 = t7 - t5 + t3 */
      paddsw   mm1,  mm4      /*P01 mm1 = t1 = t6 + t4 - t2 */
      psubsw   mm2,  mm4      /*P01 mm2 = t2 = t6 - t4 + t2 */
      
   /* calculate and store t0+t4, and t0-t4
         registers used: eax = data
                         edx = col_const
                         mm6
         registers freed: mm0 */
   /* op_code  dest        src */
      movq     mm4,        [eax+16]    /*P2    mm4 = x1 */
      movq     mm5,        [eax+48]    /*P2    mm5 = x3 */
      movq     mm6,        [eax+80]    /*P2    mm6 = x5 */
      movq     mm7,        [eax+112]   /*P2    mm7 = x7 */

      pmulhw   mm4,        [edx+32]    /*P2 P0 mm4 =  cosp1_16 * x1 */
      pmulhw   mm5,        [edx+40]    /*P2 P0 mm5 =  cosp3_16 * x3 */
      paddsw   mm4,        mm5         /*P01 */
      pmulhw   mm6,        [edx+48]    /*P2 P0 mm6 = -cosp5_16 * x5 */
      psubsw   mm4,        mm6         /*P01 */
      movq     mm6,        mm7         /*P01   mm6 = x7 */
      pmulhw   mm7,        [edx+56]    /*P2 P0 mm7 =  cosp7_16 * x7 */
      paddsw   mm4,        mm7         /*P01   mm4 = t4 */
      paddsw   mm4,        [edx+72]    /*P2 P01 reduce rounding error */
      movq     mm5,        mm4         /*P01 */
      paddsw   mm4,        mm0         /*P01   mm4 = t0+t4 */
      psubsw   mm0,        mm5         /*P01   mm0 = t0-t4 */

      psraw    mm4,        col_shift   /*P1 */
      movq     [eax],      mm4         /*P3 P4 */
      psraw    mm0,        col_shift   /*P1 */
      movq     [eax+112],  mm0         /*P3 P4 free mm0 */

   /* calculate and store t1+t5, and t1-t5
         registers used: eax = data
                         edx = col_const
                         mm5
         registers freed: mm1 */
   /* op_code  dest        src */
      movq     mm7,        mm6         /*P01   mm7 = x7 */
      movq     mm0,        [eax+80]    /*P2    mm0 = x5 */
      movq     mm5,        [eax+48]    /*P2    mm5 = x3 */
      movq     mm4,        [eax+16]    /*P2    mm4 = x1 */

      pmulhw   mm7,        [edx+48]    /*P2 P0 mm7 = -cosp5_16 * x7 */
      pmulhw   mm0,        [edx+32]    /*P2 P0 mm0 =  cosp1_16 * x5 */
      psubsw   mm7,        mm0         /*P01 */
      pmulhw   mm5,        [edx+56]    /*P2 P0 mm5 =  cosp7_16 * x3 */
      psubsw   mm7,        mm5         /*P01 */
      movq     mm5,        mm4         /*P01   mm5 = x1 */
      pmulhw   mm4,        [edx+40]    /*P2 P0 mm4 =  cosp3_16 * x1 */
      paddsw   mm7,        mm4         /*P01   mm7 = t5 */
      movq     mm0,        mm7
      paddsw   mm7,        mm1         /*P01   mm7 = t1+t5 */
      psubsw   mm1,        mm0         /*P01   mm1 = t1-t5 */

      psraw    mm7,        col_shift   /*P1 */
      movq     [eax+16],   mm7         /*P3 P4 */
      psraw    mm1,        col_shift   /*P1 */
      movq     [eax+96],   mm1         /*P3 P4 free mm1 */

   /* calculate and store t2+t6, and t2-t6
         registers used: eax = data
                         edx = col_const
                         mm1
         registers freed: mm2 */
   /* op_code  dest        src */
      movq     mm7,        mm6         /*P01   mm7 = x7 */
      movq     mm4,        mm5         /*P01   mm4 = x1 */
      movq     mm1,        [eax+48]    /*P2    mm1 = x3 */
      movq     mm0,        [eax+80]    /*P2    mm0 = x5 */

      pmulhw   mm7,        [edx+40]    /*P2 P0 mm7 =  cosp3_16 * x7 */
      pmulhw   mm4,        [edx+48]    /*P2 P0 mm4 = -cosp5_16 * x1 */
      psubsw   mm7,        mm4         /*P01 */
      pmulhw   mm1,        [edx+32]    /*P2 P0 mm1 =  cosp1_16 * x3 */
      psubsw   mm7,        mm1         /*P01 */
      movq     mm1,        mm0         /*P01   mm1 = x5 */
      pmulhw   mm0,        [edx+56]    /*P2 P0 mm0 =  cosp7_16 * x5 */
      paddsw   mm7,        mm0         /*P01   mm7 = t6 */
      movq     mm4,        mm7         /*P01 */
      paddsw   mm7,        mm2         /*P01   mm7 = t2+t6 */
      psubsw   mm2,        mm4         /*P01   mm2 = t2-t6 */

      psraw    mm7,        col_shift   /*P1 */
      movq     [eax+32],   mm7         /*P3 P4 */
      psraw    mm2,        col_shift   /*P1 */
      movq     [eax+80],   mm2         /*P3 P4 free mm2 */

   /* calculate and store t3+t7, and t3-t7
         registers used: eax = data
                         edx = col_const
         registers freed: mm1, mm3, mm5, mm6 */
   /* op_code  dest        src */
      movq     mm4,        [eax+48]    /*P2    mm4 = x3 */

      pmulhw   mm5,        [edx+56]    /*P2 P0 mm5 =  cosp7_16 * x1 */
      pmulhw   mm4,        [edx+48]    /*P2 P0 mm4 = -cosp5_16 * x3 */
      paddsw   mm5,        mm4         /*P01 */
      pmulhw   mm1,        [edx+40]    /*P2 P0 mm1 =  cosp3_16 * x5 */
      paddsw   mm5,        mm1         /*P01 */
      pmulhw   mm6,        [edx+32]    /*P2 P0 mm6 =  cosp1_16 * x7 */
      psubsw   mm5,        mm6         /*P01   mm5 = t7 */
      paddsw   mm5,        [edx+72]    /*P2 P01 reduce rounding error */
      movq     mm4,        mm5         /*P01 */
      paddsw   mm5,        mm3         /*P01   mm5 = t3+t7 */
      psubsw   mm3,        mm4         /*P01   mm3 = t3-t7 */

      psraw    mm5,        col_shift   /*P1 */
      movq     [eax+48],   mm5         /*P3 P4 */
      psraw    mm3,        col_shift   /*P1 */
      movq     [eax+64],   mm3         /*P3 P4 free mm3 */

      add      eax,        8           /*P01   data += 4; */
      dec      ecx                     /*P01   loop counter */
      jnz      IDCT_COL_LOOP
      emms
   }

#else
# error "Don't know how to insert inline ASM code for this compiler"
#endif
}

/*
IEEE test conditions: -L = -256, +H = 255, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0117   0.0134   0.0125   0.0134   0.0118   0.0138   0.0129   0.0140
   0.0127   0.0154   0.0131   0.0119   0.0112   0.0121   0.0148   0.0115
   0.0133   0.0128   0.0136   0.0108   0.0129   0.0142   0.0137   0.0118
   0.0110   0.0128   0.0116   0.0110   0.0117   0.0125   0.0120   0.0094
   0.0117   0.0121   0.0121   0.0133   0.0116   0.0133   0.0132   0.0126
   0.0129   0.0131   0.0118   0.0110   0.0105   0.0100   0.0135   0.0124
   0.0134   0.0124   0.0131   0.0119   0.0138   0.0103   0.0113   0.0131
   0.0128   0.0122   0.0105   0.0128   0.0115   0.0121   0.0117   0.0122
Worst pmse = 0.015400  (meets spec limit 0.06)
Overall mse = 0.012367  (meets spec limit 0.02)

Mean errors:
   0.0041   0.0052   0.0049   0.0066   0.0044   0.0046   0.0053   0.0050
  -0.0039  -0.0052  -0.0043  -0.0031  -0.0040  -0.0025  -0.0054  -0.0035
  -0.0015   0.0004  -0.0024  -0.0010  -0.0015   0.0002   0.0009  -0.0020
  -0.0036  -0.0020  -0.0018  -0.0008  -0.0025  -0.0037  -0.0034  -0.0010
  -0.0013  -0.0013  -0.0001  -0.0005  -0.0018  -0.0013  -0.0014   0.0002
  -0.0019  -0.0013  -0.0016  -0.0016  -0.0005  -0.0016   0.0001  -0.0034
  -0.0024  -0.0016  -0.0019  -0.0019  -0.0016  -0.0023  -0.0029  -0.0019
  -0.0022  -0.0014  -0.0011  -0.0008  -0.0009  -0.0011  -0.0025  -0.0002
Worst mean error = 0.006600  (meets spec limit 0.015)
Overall mean error = -0.000992  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero

*************************************************************************

IEEE test conditions: -L = -5, +H = 5, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0118   0.0147   0.0133   0.0126   0.0106   0.0125   0.0136   0.0112
   0.0120   0.0157   0.0129   0.0120   0.0130   0.0132   0.0141   0.0119
   0.0125   0.0132   0.0128   0.0111   0.0119   0.0122   0.0127   0.0122
   0.0109   0.0096   0.0116   0.0134   0.0132   0.0105   0.0129   0.0123
   0.0117   0.0136   0.0124   0.0118   0.0098   0.0120   0.0107   0.0116
   0.0132   0.0124   0.0132   0.0120   0.0113   0.0126   0.0115   0.0138
   0.0130   0.0116   0.0144   0.0099   0.0128   0.0129   0.0110   0.0119
   0.0113   0.0108   0.0115   0.0128   0.0130   0.0124   0.0128   0.0117
Worst pmse = 0.015700  (meets spec limit 0.06)
Overall mse = 0.012273  (meets spec limit 0.02)

Mean errors:
   0.0044   0.0073   0.0051   0.0066   0.0034   0.0033   0.0066   0.0042
  -0.0048  -0.0045  -0.0035  -0.0050  -0.0042  -0.0042  -0.0015  -0.0033
  -0.0007   0.0006   0.0006   0.0011  -0.0025  -0.0006  -0.0007   0.0014
  -0.0017  -0.0022  -0.0020  -0.0016  -0.0018  -0.0009  -0.0019  -0.0015
  -0.0007  -0.0006  -0.0018  -0.0020  -0.0012   0.0000   0.0007  -0.0018
   0.0000  -0.0016  -0.0016   0.0002  -0.0011  -0.0018  -0.0031  -0.0018
  -0.0006  -0.0018  -0.0022  -0.0003  -0.0016  -0.0019  -0.0020  -0.0005
   0.0005  -0.0018  -0.0021  -0.0014  -0.0026  -0.0012   0.0004  -0.0015
Worst mean error = 0.007300  (meets spec limit 0.015)
Overall mean error = -0.000677  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero

*************************************************************************

IEEE test conditions: -L = -300, +H = 300, sign = 1, #iters = 10000
Peak absolute values of errors:
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
   1   1   1   1   1   1   1   1
Worst peak error = 1  (meets spec limit 1)

Mean square errors:
   0.0101   0.0116   0.0105   0.0111   0.0112   0.0103   0.0117   0.0117
   0.0104   0.0105   0.0107   0.0102   0.0100   0.0121   0.0110   0.0131
   0.0098   0.0089   0.0138   0.0098   0.0106   0.0092   0.0098   0.0128
   0.0094   0.0102   0.0115   0.0122   0.0123   0.0121   0.0108   0.0089
   0.0112   0.0100   0.0099   0.0109   0.0099   0.0100   0.0104   0.0121
   0.0100   0.0119   0.0099   0.0093   0.0098   0.0112   0.0085   0.0110
   0.0118   0.0092   0.0094   0.0106   0.0113   0.0113   0.0102   0.0096
   0.0112   0.0115   0.0107   0.0097   0.0117   0.0106   0.0099   0.0112
Worst pmse = 0.013800  (meets spec limit 0.06)
Overall mse = 0.010691  (meets spec limit 0.02)

Mean errors:
   0.0035   0.0042   0.0043   0.0039   0.0028   0.0049   0.0037   0.0039
  -0.0030  -0.0025  -0.0025  -0.0036  -0.0036  -0.0039  -0.0024  -0.0041
   0.0006  -0.0019   0.0028  -0.0012   0.0004  -0.0010   0.0000   0.0000
  -0.0024  -0.0022  -0.0019  -0.0010  -0.0009  -0.0021  -0.0028  -0.0029
  -0.0008  -0.0014   0.0005  -0.0011  -0.0017  -0.0014  -0.0014  -0.0011
  -0.0010  -0.0027  -0.0011  -0.0019  -0.0024  -0.0016  -0.0013  -0.0034
  -0.0022  -0.0014  -0.0008  -0.0010  -0.0017  -0.0015  -0.0002  -0.0012
  -0.0012  -0.0023  -0.0023  -0.0013  -0.0027  -0.0004  -0.0013  -0.0008
Worst mean error = 0.004900  (meets spec limit 0.015)
Overall mean error = -0.000891  (meets spec limit 0.0015)

0 elements of IDCT(0) were not zero
*/

#undef shift_value
#undef row_shift
#undef row_error
#undef col_shift
#undef col_error
#undef cosp1_16
#undef cosp2_16
#undef cosp3_16
#undef cosp4_16
#undef cosp5_16
#undef cosp6_16
#undef cosp7_16
#undef cosm1_16
#undef cosm2_16
#undef cosm3_16
#undef cosm4_16
#undef cosm5_16
#undef cosm6_16
#undef cosm7_16
#undef r_const
#undef c_const

